"""
pglookout tests

Copyright (c) 2015 Ohmu Ltd
Copyright (c) 2014 F-Secure

This file is under the Apache License, Version 2.0.
See the file `LICENSE` for details.
"""

from pglookout.common import get_iso_timestamp
from pglookout.pglookout import PgLookout
from pglookout.pgutil import get_connection_info, get_connection_info_from_config_line
try:
    from mock import Mock  # pylint: disable=import-error
except ImportError:  # py3k import location
    from unittest.mock import Mock  # pylint: disable=import-error,no-name-in-module
from unittest import TestCase
import datetime
import os
import shutil
import tempfile
import time


def _create_db_node_state(pg_last_xlog_receive_location=None, pg_is_in_recovery=True,
                          connection=True, replication_time_lag=None, fetch_time=None,
                          db_time=None):
    return {
        "connection": connection,
        "db_time": get_iso_timestamp(db_time),
        "fetch_time": get_iso_timestamp(fetch_time),
        "pg_is_in_recovery": pg_is_in_recovery,
        "pg_last_xact_replay_timestamp": None,
        "pg_last_xlog_receive_location": pg_last_xlog_receive_location,
        "pg_last_xlog_replay_location": None,
        "replication_time_lag": replication_time_lag,
        "min_replication_time_lag": 0,  # simulate that we've been in sync once
    }


class TestClusterMonitor(TestCase):
    def setUp(self):
        self.pglookout = PgLookout("pglookout.json")
        self.original_connect_to_db = self.pglookout.cluster_monitor._connect_to_db   # pylint: disable=protected-access
        self.pglookout.cluster_monitor._connect_to_db = Mock()  # pylint: disable=protected-access

    def test_connect_to_cluster_nodes_and_cleanup_old_nodes(self):
        self.pglookout.cluster_monitor.db_conns = {"1.2.3.4": "bar", "2.3.4.5": "foo", "3.4.5.6": "foo", None: "foo"}
        self.pglookout.cluster_monitor.connect_to_cluster_nodes_and_cleanup_old_nodes()
        self.assertEqual(self.pglookout.cluster_monitor.db_conns, {"1.2.3.4": "bar", "2.3.4.5": "foo"})

    def tearDown(self):
        self.pglookout.cluster_monitor._connect_to_db = self.original_connect_to_db   # pylint: disable=protected-access


class TestPgLookout(TestCase):
    def setUp(self):
        self.pglookout = PgLookout("pglookout.json")
        self.pglookout.execute_external_command = Mock()
        self.pglookout.create_alert_file = Mock()
        self.pglookout.check_for_maintenance_mode_file = Mock()
        self.pglookout.check_for_maintenance_mode_file.return_value = False
        self.temp_dir = tempfile.mkdtemp(prefix="pglookout_test_")
        self.state_file_path = os.path.join(self.temp_dir, "state_file")

    def tearDown(self):
        if os.path.exists(self.temp_dir) and self.temp_dir.startswith("/tmp/pglookout_test_"):
            shutil.rmtree(self.temp_dir)

    def test_state_file_write(self):
        self.pglookout.config['json_state_file_path'] = self.state_file_path
        self.pglookout.write_cluster_state_to_json_file()
        self.assertTrue(os.path.exists(self.state_file_path))
        self.assertTrue(os.path.getsize(self.state_file_path), 2)
        os.unlink(self.state_file_path)

    def test_load_config(self):
        self.pglookout.own_db = "old_value"
        self.pglookout.load_config()
        self.assertEqual(self.pglookout.own_db, "1.2.3.4")

    def _add_to_observer_state(self, observer_name, db_name, pg_last_xlog_receive_location=None,
                               pg_is_in_recovery=True, connection=True, replication_time_lag=None,
                               fetch_time=None, db_time=None):
        db_node_state = _create_db_node_state(pg_last_xlog_receive_location, pg_is_in_recovery,
                                              connection, replication_time_lag, fetch_time=fetch_time,
                                              db_time=db_time)
        update_dict = {"fetch_time": get_iso_timestamp(),
                       "connection": True, db_name: db_node_state}
        if observer_name in self.pglookout.observer_state:
            self.pglookout.observer_state[observer_name].update(update_dict)
        else:
            self.pglookout.observer_state[observer_name] = update_dict

    def _add_db_to_cluster_state(self, instance, pg_last_xlog_receive_location=None,
                                 pg_is_in_recovery=True, connection=True, replication_time_lag=None,
                                 fetch_time=None, db_time=None, conn_info=None):
        db_node_state = _create_db_node_state(pg_last_xlog_receive_location, pg_is_in_recovery,
                                              connection, replication_time_lag, fetch_time=fetch_time,
                                              db_time=db_time)
        self.pglookout.cluster_state[instance] = db_node_state
        self.pglookout.config["remote_conns"][instance] = conn_info or {"host": instance}

    def test_check_cluster_state_warning(self):
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=40.0)
        self.pglookout.own_db = "kuu"
        self.pglookout.over_warning_limit_command = "fake_command"
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)
        self.assertEqual(self.pglookout.create_alert_file.call_count, 1)
        self.pglookout.check_cluster_state()

        # call count does not change when we have sent a single warning
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)
        self.assertEqual(self.pglookout.create_alert_file.call_count, 1)

        # and then the replication catches up
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=5.0)
        self.pglookout.check_cluster_state()
        self.assertFalse(os.path.exists("replication_delay_warning"))
        self.assertFalse(self.pglookout.replication_lag_over_warning_limit)

    def test_check_cluster_do_failover_one_slave(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))

        self._add_db_to_cluster_state("own_db", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)

        self.pglookout.own_db = "own_db"
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = False
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)
        self.assertFalse(self.pglookout.replication_lag_over_warning_limit)

    def test_check_cluster_do_failover_one_slave_one_observer(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))

        self._add_db_to_cluster_state("own_db", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "own_db"
        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=False,
                                    db_time=datetime.datetime(year=2014, month=1, day=1))
        self._add_to_observer_state("observer", "own_db", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)

        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = False
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)
        self.assertFalse(self.pglookout.replication_lag_over_warning_limit)

    def test_check_cluster_do_failover_with_a_node_which_is_is_maintenance(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))

        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)

        self.pglookout.never_promote_these_nodes = []
        self.pglookout.own_db = "kuu"
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = True
        self.pglookout.check_for_maintenance_mode_file.return_value = True
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)
        self.assertEqual(self.pglookout.check_for_maintenance_mode_file.call_count, 1)

    def test_check_cluster_do_failover_with_a_node_which_should_never_be_promoted(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False)

        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.never_promote_these_nodes = ["kuu"]
        self.pglookout.own_db = "kuu"
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = True
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)

    def test_check_cluster_do_failover_two_slaves(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False)

        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "kuu"
        # we put the second slave _WELL_ ahead
        self._add_db_to_cluster_state("puu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)

        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = True
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

    def test_check_cluster_do_failover_two_slaves_when_the_one_ahead_can_never_be_promoted(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))

        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "kuu"
        # we put the second slave _WELL_ ahead
        self._add_db_to_cluster_state("puu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.never_promote_these_nodes = ["puu"]
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.replication_lag_over_warning_limit = True
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)
        self.assertFalse(self.pglookout.replication_lag_over_warning_limit)

    def test_failover_with_no_master_anymore(self):
        # this should not trigger an immediate failover as we have two
        # standbys online but we've never seen a master so we wait a while
        # and see what happens
        self.pglookout.own_db = "kuu"
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="F/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=0)
        self._add_db_to_cluster_state("puu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=1)

        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.check_cluster_state()
        assert self.pglookout.execute_external_command.call_count == 0

        # now we add a fake "current" master indicating that the cluster has
        # been consistent at some point, this should trigger an immediate
        # failover
        self.pglookout.current_master = "something obsolete"
        self.pglookout.check_cluster_state()
        # No failover yet since we're  not over missing_master_from_config_timeout
        assert self.pglookout.execute_external_command.call_count == 0

        self.pglookout.cluster_nodes_change_time = time.time() - self.pglookout.missing_master_from_config_timeout
        self.pglookout.current_master = "something obsolete"
        self.pglookout.check_cluster_state()
        assert self.pglookout.execute_external_command.call_count == 1

    def test_failover_with_no_master_timeout(self):
        # this should not trigger an immediate failover as we have two
        # standbys online but we've never seen a master so we wait a while
        # and see what happens
        self.pglookout.own_db = "kuu"
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="F/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=0)
        self._add_db_to_cluster_state("puu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=1)

        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.check_cluster_state()
        assert self.pglookout.execute_external_command.call_count == 0

        # indicate that we haven't seen configuration changes for 5 minutes,
        # that should trigger a failover as the timeout has passed
        self.pglookout.cluster_nodes_change_time = time.time() - 300
        self.pglookout.check_cluster_state()
        assert self.pglookout.execute_external_command.call_count == 1

    def test_failover_over_replication_lag_when_still_connected_to_master(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False)

        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "kuu"

        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

    def test_failover_over_replication_lag_with_one_observer_one_slave_no_connections(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False)

        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("own_db", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "own_db"

        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=False,
                                    db_time=datetime.datetime(year=2014, month=1, day=1))
        self._add_to_observer_state("observer", "own_db", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)
        self.pglookout.observer_state["observer"]['connection'] = False
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

    def test_cluster_state_when_observer_has_also_non_members_of_our_current_cluster(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=True)

        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("own_db", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "own_db"

        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=False,
                                    db_time=datetime.datetime(year=2014, month=1, day=1))
        self._add_to_observer_state("observer", "own_db", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)
        self._add_to_observer_state("observer", "some_other_cluster", pg_last_xlog_receive_location="3/aaaaaaaa",
                                    pg_is_in_recovery=False, connection=True, replication_time_lag=0.0)
        self.pglookout.check_cluster_state()
        self.assertEqual(len(self.pglookout.connected_master_nodes), 1)
        assert 'old_master' in self.pglookout.connected_master_nodes

    def test_failover_no_connections(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False)

        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("kuu", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "kuu"

        # we put the second slave _WELL_ ahead
        self._add_db_to_cluster_state("puu", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

    def test_failover_master_two_slaves_one_observer_no_connection_between_slaves(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))
        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.own_db = "own"

        self._add_db_to_cluster_state("other", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)

        # Add observer state
        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=False,
                                    db_time=datetime.datetime(year=2014, month=1, day=1))
        self._add_to_observer_state("observer", "other", pg_last_xlog_receive_location="1/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self._add_to_observer_state("observer", "own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=True, replication_time_lag=130.0)
        self.pglookout.execute_external_command.return_value = 0
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)

        self.assertFalse(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

    def test_failover_master_one_slave_one_observer_no_connections(self):
        self.pglookout.own_db = "own"

        # Add observer state
        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=True)

        # add db state
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=True)
        self._add_db_to_cluster_state("own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=40.0)

        self.pglookout.check_cluster_state()
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)

        # Add observer state
        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=True)
        self._add_to_observer_state("observer", "own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=True, replication_time_lag=9.0)

        self._add_db_to_cluster_state("own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=140.0)

        self.pglookout.check_cluster_state()

        # No failover yet
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on

        # observer state
        self._add_to_observer_state("observer", "old_master", pg_is_in_recovery=False, connection=False,
                                    db_time=datetime.datetime(year=2014, month=1, day=1))
        self._add_to_observer_state("observer", "own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                    pg_is_in_recovery=True, connection=False, replication_time_lag=140.0)
        # lose own connection to master
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      db_time=datetime.datetime(year=2014, month=1, day=1))
        # now do failover
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)

    def test_find_current_master(self):
        self._add_db_to_cluster_state("master", pg_is_in_recovery=False, connection=True)
        # We will make our own node to be the furthest along so we get considered for promotion
        self._add_db_to_cluster_state("own", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=True, replication_time_lag=0.1)
        self.pglookout.own_db = "master"
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.current_master, "master")

    def test_two_slave_failover_and_autofollow(self):
        self._add_db_to_cluster_state("old_master", pg_is_in_recovery=False, connection=False,
                                      fetch_time=datetime.datetime(year=2014, month=1, day=1))
        # We will make our own node to be the furthest from master so we don't get considered for promotion
        self._add_db_to_cluster_state("own", pg_last_xlog_receive_location="1/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)
        self.pglookout.own_db = "own"
        self._add_db_to_cluster_state("other", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=True, connection=False, replication_time_lag=130.0)
        self.pglookout.check_cluster_state()

        self.assertTrue(self.pglookout.replication_lag_over_warning_limit)  # we keep the warning on
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        self.assertEqual(self.pglookout.current_master, "old_master")

        self._add_db_to_cluster_state("other", pg_last_xlog_receive_location="2/aaaaaaaa",
                                      pg_is_in_recovery=False, connection=True, replication_time_lag=0.0,
                                      conn_info={"host": "otherhost.example.com", "port": 11111})

        pg_data_dir = os.path.join(self.temp_dir + os.sep + "test_pgdata")
        os.makedirs(pg_data_dir)
        primary_conninfo = (
            "user=replication password=vjsh8l7sv4a902y1tsdz "
            "host=old_master port=5432 "
            "sslmode=prefer sslcompression=1 krbsrvname=postgres"
        )
        old_recovery_conf = "standby_mode = 'on'\nprimary_conninfo = '{0}'\n".format(primary_conninfo)
        with open(os.path.join(pg_data_dir, "recovery.conf"), "w") as fp:
            fp.write(old_recovery_conf)

        self.pglookout.config['pg_data_directory'] = pg_data_dir
        self.pglookout.config['autofollow'] = True
        self.pglookout.primary_conninfo_template = get_connection_info(primary_conninfo)

        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.current_master, "other")

        with open(os.path.join(pg_data_dir, "recovery.conf"), "r") as fp:
            new_lines = fp.read().splitlines()
        assert new_lines.pop(0).startswith("# pglookout updated primary_conninfo")
        assert new_lines.pop(0) == "standby_mode = 'on'"
        assert new_lines[0].startswith("primary_conninfo = ")
        new_primary_conninfo = new_lines.pop(0)
        assert new_lines.pop(0) == "recovery_target_timeline = 'latest'"
        assert new_lines == []
        old_conn_info = get_connection_info(primary_conninfo)
        new_conn_info = get_connection_info_from_config_line(new_primary_conninfo)
        assert new_conn_info == dict(old_conn_info, host="otherhost.example.com", port="11111")

    def test_replication_positions(self):
        standby_nodes = {
            '10.255.255.10': {
                'connection': True,
                'db_time': '2014-08-28T14:09:57.919301+00:00Z',
                'fetch_time': '2014-08-28T14:09:57.918753Z',
                'pg_is_in_recovery': True,
                'pg_last_xlog_receive_location': '0/9000090',
                'pg_last_xlog_replay_location': '0/9000090',
                'pg_last_xact_replay_timestamp': '2014-08-28T14:05:43.577357+00:00Z',
                'replication_time_lag': 254.341944,
            },
        }
        # the above node shouldn't show up as it's fetch_time is (way) older than 20 seconds
        positions = {}
        assert self.pglookout.get_replication_positions(standby_nodes) == positions
        standby_nodes['10.255.255.10']['fetch_time'] = get_iso_timestamp()
        positions[0x9000090] = set(['10.255.255.10'])
        assert self.pglookout.get_replication_positions(standby_nodes) == positions
        # add another standby, further ahead
        standby_nodes['10.255.255.11'] = dict(standby_nodes['10.255.255.10'], pg_last_xlog_receive_location='1/0000AAAA')
        positions[1 << 32 | 0xAAAA] = set(['10.255.255.11'])
        assert self.pglookout.get_replication_positions(standby_nodes) == positions
        # add another standby which hasn't received anything
        standby_nodes['10.255.255.12'] = dict(standby_nodes['10.255.255.10'], pg_last_xlog_receive_location=None)
        positions[0x9000090].add('10.255.255.12')
        assert self.pglookout.get_replication_positions(standby_nodes) == positions

    def test_node_map(self):
        cluster_state = {
            "10.255.255.10": {
                "connection": True,
                "db_time": "2014-08-28T14:26:51.067084+00:00Z",
                "fetch_time": "2014-08-28T14:26:51.066368Z",
                "pg_is_in_recovery": False,
                "pg_last_xact_replay_timestamp": "2014-08-28T14:05:43.577357+00:00Z",
                "pg_last_xlog_receive_location": "0/9000090",
                "pg_last_xlog_replay_location": "0/9000090",
                "replication_time_lag": 1267.489727,
            },
            "10.255.255.9": {
                "connection": False,
                "fetch_time": "2014-08-28T14:26:51.068151Z",
            }
        }
        observer_state = {
            "10.255.255.11": {
                "10.255.255.10": {
                    "connection": True,
                    "db_time": "2014-08-28T14:26:47.105901+00:00Z",
                    "fetch_time": "2014-08-28T14:26:47.104849Z",
                    "pg_is_in_recovery": False,
                    "pg_last_xact_replay_timestamp": "2014-08-28T14:05:43.577357+00:00Z",
                    "pg_last_xlog_receive_location": "0/9000090",
                    "pg_last_xlog_replay_location": "0/9000090",
                    "replication_time_lag": 1263.528544,
                },
                "10.255.255.9": {
                    "connection": False,
                    "db_time": "2014-08-28T14:06:15.172820+00:00Z",
                    "fetch_time": "2014-08-28T14:26:47.107115Z",
                    "pg_is_in_recovery": False,
                    "pg_last_xact_replay_timestamp": None,
                    "pg_last_xlog_receive_location": None,
                    "pg_last_xlog_replay_location": None,
                },
                "connection": True,
                "fetch_time": "2014-08-28T14:26:51.069891Z",
            }
        }
        master_host, _, standby_nodes = self.pglookout.create_node_map(cluster_state, observer_state)
        self.assertEqual(master_host, "10.255.255.10")
        self.assertEqual(standby_nodes, {})

    def test_node_map_disconnected_current_master(self):
        self.pglookout.current_master = "10.255.255.7"
        cluster_state = {
            "10.255.255.7": {
                "connection": False,
                "db_time": "2014-09-07T15:26:23.957151+00:00Z",
                "fetch_time": "2014-09-07T15:26:34.736495Z",
                "pg_is_in_recovery": False,
                "pg_last_xact_replay_timestamp": None,
                "pg_last_xlog_receive_location": None,
                "pg_last_xlog_replay_location": None,
            },
            "10.255.255.8": {
                "connection": True,
                "db_time": "2014-09-07T15:26:23.959461+00:00Z",
                "fetch_time": "2014-09-07T15:26:23.919281Z",
                "pg_is_in_recovery": True,
                "pg_last_xact_replay_timestamp": "2014-09-07T15:25:40.372936+00:00Z",
                "pg_last_xlog_receive_location": "0/74713D8",
                "pg_last_xlog_replay_location": "0/74713D8",
                "replication_time_lag": 43.586525,
            }
        }
        observer_state = {}
        master_host, _, standby_nodes = self.pglookout.create_node_map(cluster_state, observer_state)
        self.assertEqual(master_host, "10.255.255.7")
        self.assertEqual(list(standby_nodes.keys())[0], "10.255.255.8")

    def test_standbys_failover_equal_replication_positions(self):
        now = get_iso_timestamp(datetime.datetime.utcnow())
        self.pglookout.cluster_state = {
            "192.168.54.183": {
                "connection": True,
                "db_time": now,
                "fetch_time": now,
                "pg_is_in_recovery": True,
                "pg_last_xact_replay_timestamp": "2015-04-28T11:21:56.098946+00:00Z",
                "pg_last_xlog_receive_location": "0/70004D8",
                "pg_last_xlog_replay_location": "0/70004D8",
                "replication_time_lag": 400.435871,
                "min_replication_time_lag": 0,  # simulate that we've been in sync once
            },
            "192.168.57.180": {
                "connection": False,
                "db_time": "2015-04-28T11:21:55.830432Z",
                "fetch_time": now,
                "pg_is_in_recovery": False,
                "pg_last_xact_replay_timestamp": None,
                "pg_last_xlog_receive_location": None,
                "pg_last_xlog_replay_location": None,
                "replication_time_lag": 0.0,
                "min_replication_time_lag": 0,  # simulate that we've been in sync once
            },
            "192.168.63.4": {
                "connection": True,
                "db_time": now,
                "fetch_time": now,
                "pg_is_in_recovery": True,
                "pg_last_xact_replay_timestamp": "2015-04-28T11:21:56.098946+00:00Z",
                "pg_last_xlog_receive_location": "0/70004D8",
                "pg_last_xlog_replay_location": "0/70004D8",
                "replication_time_lag": 401.104655,
                "min_replication_time_lag": 0,  # simulate that we've been in sync once
            },
        }
        self.pglookout.current_master = "192.168.57.180"
        # We select the node with the "highest" identifier so call_count should stay zero if we're not the
        # highest standby currently.
        self.pglookout.own_db = "192.168.54.183"
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 0)
        # If we're the highest we should see call_count increment
        self.pglookout.own_db = "192.168.63.4"
        self.pglookout.check_cluster_state()
        self.assertEqual(self.pglookout.execute_external_command.call_count, 1)



"""
pglookout - webserver component

Copyright (c) 2015 Ohmu Ltd
Copyright (c) 2014 F-Secure

This file is under the Apache License, Version 2.0.
See the file `LICENSE` for details.
"""
import json
from logging import getLogger
from threading import Thread

try:
    from SocketServer import ThreadingMixIn  # pylint: disable=import-error
    from BaseHTTPServer import HTTPServer  # pylint: disable=import-error
    from SimpleHTTPServer import SimpleHTTPRequestHandler  # pylint: disable=import-error
except ImportError:  # Support Py3k
    from socketserver import ThreadingMixIn  # pylint: disable=import-error
    from http.server import HTTPServer, SimpleHTTPRequestHandler  # pylint: disable=import-error


class ThreadedWebServer(ThreadingMixIn, HTTPServer):
    cluster_state = None
    log = None


class WebServer(Thread):
    def __init__(self, config, cluster_state):
        Thread.__init__(self)
        self.config = config
        self.cluster_state = cluster_state
        self.log = getLogger("WebServer")
        self.address = self.config.get("http_address", '')
        self.port = self.config.get("http_port", 15000)
        self.server = None
        self.log.debug("WebServer initialized with address: %r port: %r", self.address, self.port)

    def run(self):
        # We bind the port only when we start running
        self.server = ThreadedWebServer((self.address, self.port), RequestHandler)
        self.server.cluster_state = self.cluster_state
        self.server.log = self.log
        self.server.serve_forever()

    def close(self):
        self.log.debug("Closing WebServer")
        self.server.shutdown()
        self.log.debug("Closed WebServer")


class RequestHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        self.server.log.debug("Got request: %r", self.path)
        if self.path.startswith("/state.json"):
            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            response = json.dumps(self.server.cluster_state, indent=4)
            self.send_header('Content-length', len(response))
            self.end_headers()
            self.wfile.write(response)
        else:
            self.send_response(404)



# Copied from https://github.com/ohmu/ohmu_common_py ohmu_common_py/statsd.py version 0.0.1-0-unknown-b16ec0a
"""
pglookout - StatsD client

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details

Supports telegraf's statsd protocol extension for 'key=value' tags:

    https://github.com/influxdata/telegraf/tree/master/plugins/inputs/statsd
"""

import logging
import socket


class StatsClient(object):
    def __init__(self, host="127.0.0.1", port=8125, tags=None):
        self.log = logging.getLogger("StatsClient")
        self._dest_addr = (host, port)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._tags = tags or {}

    def gauge(self, metric, value, tags=None):
        self._send(metric, b"g", value, tags)

    def increase(self, metric, inc_value=1, tags=None):
        self._send(metric, b"c", inc_value, tags)

    def timing(self, metric, value, tags=None):
        self._send(metric, b"ms", value, tags)

    def unexpected_exception(self, ex, where, tags=None):
        all_tags = {
            "exception": ex.__class__.__name__,
            "where": where,
        }
        all_tags.update(tags or {})
        self.increase("exception", tags=all_tags)

    def _send(self, metric, metric_type, value, tags):
        if None in self._dest_addr:
            # stats sending is disabled
            return

        try:
            # format: "user.logins,service=payroll,region=us-west:1|c"
            parts = [metric.encode("utf-8"), b":", str(value).encode("utf-8"), b"|", metric_type]
            send_tags = self._tags.copy()
            send_tags.update(tags or {})
            for tag, value in send_tags.items():
                parts.insert(1, ",{}={}".format(tag, value).encode("utf-8"))

            self._socket.sendto(b"".join(parts), self._dest_addr)
        except Exception as ex:  # pylint: disable=broad-except
            self.log.error("Unexpected exception in statsd send: %s: %s",
                           ex.__class__.__name__, ex)



"""
pglookout - replication monitoring and failover daemon

Copyright (c) 2015 Ohmu Ltd
Copyright (c) 2014 F-Secure

This file is under the Apache License, Version 2.0.
See the file `LICENSE` for details.
"""

from __future__ import print_function
from . import logutil, statsd, version
from .cluster_monitor import ClusterMonitor
from .common import convert_xlog_location_to_offset, parse_iso_datetime, get_iso_timestamp
from .pgutil import (
    create_connection_string, get_connection_info, get_connection_info_from_config_line)
from .webserver import WebServer
from psycopg2.extensions import adapt
import argparse
import copy
import datetime
import json
import logging
import logging.handlers
import os
import signal
import socket
import subprocess
import sys
import time

try:
    from queue import Queue  # pylint: disable=import-error
except ImportError:
    from Queue import Queue  # pylint: disable=import-error


class PgLookout(object):
    def __init__(self, config_path):
        self.log = logging.getLogger("pglookout")
        self.stats = None
        self.running = True
        self.replication_lag_over_warning_limit = False

        self.config_path = config_path
        self.config = {}
        self.log_level = "DEBUG"

        self.connected_master_nodes = {}
        self.disconnected_master_nodes = {}
        self.connected_observer_nodes = {}
        self.disconnected_observer_nodes = {}
        self.replication_lag_warning_boundary = None
        self.replication_lag_failover_timeout = None
        self.missing_master_from_config_timeout = None
        self.own_db = None
        self.current_master = None
        self.failover_command = None
        self.over_warning_limit_command = None
        self.never_promote_these_nodes = None
        self.primary_conninfo_template = None
        self.cluster_monitor = None
        self.syslog_handler = None
        self.cluster_nodes_change_time = time.time()
        self.trigger_check_queue = Queue()
        self.load_config()

        signal.signal(signal.SIGHUP, self.load_config)
        signal.signal(signal.SIGINT, self.quit)
        signal.signal(signal.SIGTERM, self.quit)

        self.cluster_state = {}
        self.observer_state = {}
        self.overall_state = {"db_nodes": self.cluster_state, "observer_nodes": self.observer_state,
                              "current_master": self.current_master,
                              "replication_lag_over_warning": self.replication_lag_over_warning_limit}

        self.cluster_monitor = ClusterMonitor(
            config=self.config,
            cluster_state=self.cluster_state,
            observer_state=self.observer_state,
            create_alert_file=self.create_alert_file,
            trigger_check_queue=self.trigger_check_queue,
            stats=self.stats)
        # cluster_monitor doesn't exist at the time of reading the config initially
        self.cluster_monitor.log.setLevel(self.log_level)
        self.webserver = WebServer(self.config, self.cluster_state)

        logutil.notify_systemd("READY=1")
        self.log.info("PGLookout initialized, local hostname: %r, own_db: %r, cwd: %r",
                      socket.gethostname(), self.own_db, os.getcwd())

    def quit(self, _signal=None, _frame=None):
        self.log.warning("Quitting, signal: %r, frame: %r", _signal, _frame)
        self.cluster_monitor.running = False
        self.running = False
        self.webserver.close()

    def load_config(self, _signal=None, _frame=None):
        self.log.debug("Loading JSON config from: %r, signal: %r, frame: %r",
                       self.config_path, _signal, _frame)

        previous_remote_conns = self.config.get("remote_conns")
        try:
            with open(self.config_path) as fp:
                self.config = json.load(fp)
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Invalid JSON config, exiting")
            self.stats.unexpected_exception(ex, where="load_config")
            sys.exit(1)

        # statsd settings may have changed
        stats = self.config.get("statsd", {})
        self.stats = statsd.StatsClient(host=stats.get("host"), port=stats.get("port"),
                                        tags=stats.get("tags"))

        if previous_remote_conns != self.config.get("remote_conns"):
            self.cluster_nodes_change_time = time.time()

        if self.config.get("autofollow"):
            try:
                self.primary_conninfo_template = get_connection_info(self.config["primary_conninfo_template"])
            except (KeyError, ValueError):
                self.log.exception("Invalid or missing primary_conninfo_template; not enabling autofollow")
                self.config["autofollow"] = False

        if self.cluster_monitor:
            self.cluster_monitor.config = copy.deepcopy(self.config)

        if self.config.get("syslog") and not self.syslog_handler:
            self.syslog_handler = logutil.set_syslog_handler(
                address=self.config.get("syslog_address", "/dev/log"),
                facility=self.config.get("syslog_facility", "local2"),
                logger=logging.getLogger(),
            )
        self.own_db = self.config.get("own_db")

        log_level_name = self.config.get("log_level", "DEBUG")
        self.log_level = getattr(logging, log_level_name)
        try:
            self.log.setLevel(self.log_level)
            if self.cluster_monitor:
                self.cluster_monitor.log.setLevel(self.log_level)
        except ValueError:
            print("Problem setting log level %r" % self.log_level)
            self.log.exception("Problem with log_level: %r", self.log_level)
        self.never_promote_these_nodes = self.config.get("never_promote_these_nodes", [])
        # we need the failover_command to be converted into subprocess [] format
        self.failover_command = self.config.get("failover_command", "").split()
        self.over_warning_limit_command = self.config.get("over_warning_limit_command")
        self.replication_lag_warning_boundary = self.config.get("warning_replication_time_lag", 30.0)
        self.replication_lag_failover_timeout = self.config.get("max_failover_replication_time_lag", 120.0)
        self.replication_catchup_timeout = self.config.get("replication_catchup_timeout", 300.0)
        self.missing_master_from_config_timeout = self.config.get("missing_master_from_config_timeout", 15.0)
        self.log.debug("Loaded config: %r from: %r", self.config, self.config_path)
        self.trigger_check_queue.put("new config came, recheck")

    def write_cluster_state_to_json_file(self):
        """Periodically write a JSON state file to disk"""
        start_time = time.time()
        state_file_path = self.config.get("json_state_file_path", "/tmp/pglookout_state.json")
        try:
            self.overall_state = {"db_nodes": self.cluster_state, "observer_nodes": self.observer_state,
                                  "current_master": self.current_master}
            json_to_dump = json.dumps(self.overall_state, indent=4)
            self.log.debug("Writing JSON state file to: %r, file_size: %r", state_file_path, len(json_to_dump))
            with open(state_file_path + ".tmp", "w") as fp:
                fp.write(json_to_dump)
            os.rename(state_file_path + ".tmp", state_file_path)
            self.log.debug("Wrote JSON state file to disk, took %.4fs", time.time() - start_time)
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Problem in writing JSON: %r file to disk, took %.4fs",
                               self.overall_state, time.time() - start_time)
            self.stats.unexpected_exception(ex, where="write_cluster_state_to_json_file")

    def create_node_map(self, cluster_state, observer_state):
        standby_nodes, master_node, master_instance = {}, None, None
        connected_master_nodes, disconnected_master_nodes = {}, {}
        connected_observer_nodes, disconnected_observer_nodes = {}, {}
        self.log.debug("Creating node map out of cluster_state: %r and observer_state: %r",
                       cluster_state, observer_state)
        for instance, state in cluster_state.items():
            if 'pg_is_in_recovery' in state:
                if state['pg_is_in_recovery']:
                    standby_nodes[instance] = state
                elif state['connection']:
                    connected_master_nodes[instance] = state
                elif not state['connection']:
                    disconnected_master_nodes[instance] = state
            else:
                self.log.debug("No knowledge on instance: %r state: %r of whether it's in recovery or not", instance, state)

        for observer_name, state in observer_state.items():
            connected = state.get("connection", False)
            if connected:
                connected_observer_nodes[observer_name] = state.get("fetch_time")
            else:
                disconnected_observer_nodes[observer_name] = state.get("fetch_time")
            for instance, db_state in state.items():
                if instance not in cluster_state:
                    # A single observer can observe multiple different replication clusters.
                    # Ignore data on nodes that don't belong in our own cluster
                    self.log.debug("Ignoring instance: %r since it does not belong into our own replication cluster",
                                   instance)
                    continue
                if isinstance(db_state, dict):  # other keys are "connection" and "fetch_time"
                    own_fetch_time = parse_iso_datetime(cluster_state[instance]["fetch_time"])
                    observer_fetch_time = parse_iso_datetime(db_state['fetch_time'])
                    self.log.debug("observer_name: %r, instance: %r, state: %r, observer_fetch_time: %r",
                                   observer_name, instance, db_state, observer_fetch_time)
                    if 'pg_is_in_recovery' in db_state:
                        if db_state['pg_is_in_recovery']:
                            # we always trust ourselves the most for localhost, and
                            # in case we are actually connected to the other node
                            if observer_fetch_time >= own_fetch_time and instance != self.own_db:
                                if instance not in standby_nodes or standby_nodes[instance]["connection"] is False:
                                    standby_nodes[instance] = db_state
                        else:
                            master_node = connected_master_nodes.get(instance, {})
                            connected = master_node.get("connection", False)
                            self.log.debug("Observer: %r sees %r as master, we see: %r, same_master: %r, connection: %r",
                                           observer_name, instance, self.current_master, instance == self.current_master,
                                           db_state.get('connection'))
                            if observer_fetch_time >= own_fetch_time and instance != self.own_db:
                                if connected:
                                    connected_master_nodes[instance] = db_state
                                else:
                                    disconnected_master_nodes[instance] = db_state
                    else:
                        self.log.warning("No knowledge on %r %r from observer: %r is in recovery",
                                         instance, db_state, observer_name)

        self.connected_master_nodes = connected_master_nodes
        self.disconnected_master_nodes = disconnected_master_nodes
        self.connected_observer_nodes = connected_observer_nodes
        self.disconnected_observer_nodes = disconnected_observer_nodes

        if len(self.connected_master_nodes) == 0:
            self.log.warning("No known master node, disconnected masters: %r", list(disconnected_master_nodes))
            if len(disconnected_master_nodes) > 0:
                master_instance, master_node = list(disconnected_master_nodes.items())[0]
        elif len(self.connected_master_nodes) == 1:
            master_instance, master_node = list(connected_master_nodes.items())[0]
            if disconnected_master_nodes:
                self.log.warning("Picked %r as master since %r are in a disconnected state",
                                 master_instance, disconnected_master_nodes)
        else:
            self.create_alert_file("multiple_master_warning")
            self.log.error("More than one master node connected_master_nodes: %r, disconnected_master_nodes: %r",
                           connected_master_nodes, disconnected_master_nodes)

        return master_instance, master_node, standby_nodes

    def is_restoring_or_catching_up_normally(self, state):
        """
        Return True if node is still in the replication catchup phase and
        replication lag alerts/metrics should not yet be generated.
        """
        replication_start_time = state.get("replication_start_time")
        if replication_start_time:
            replication_total_time = time.time() - replication_start_time
            if replication_total_time > self.replication_catchup_timeout:
                # we've been replicating for too long and should have caught up with the master already
                return False

        min_lag = state.get("min_replication_time_lag", self.replication_lag_warning_boundary)
        if not state.get("pg_last_xlog_receive_location"):
            # node has not received anything from the master yet
            return True
        elif min_lag >= self.replication_lag_warning_boundary:
            # node is catching up the master and has not gotten close enough yet
            return True

        # node has caught up with the master so we should be in sync
        return False

    def emit_stats(self, state):
        if self.is_restoring_or_catching_up_normally(state):
            # do not emit misleading lag stats during catchup at restore
            return

        replication_time_lag = state.get("replication_time_lag")
        if replication_time_lag is not None:
            self.stats.gauge("pg.replication_lag", replication_time_lag)

    def check_cluster_state(self):
        master_node = None
        cluster_state = copy.deepcopy(self.cluster_state)
        observer_state = copy.deepcopy(self.observer_state)
        if not cluster_state:
            self.log.warning("No cluster state, probably still starting up")
            return

        master_instance, master_node, standby_nodes = self.create_node_map(cluster_state, observer_state)

        if master_instance and master_instance != self.current_master:
            self.log.info("New master node detected: old: %r new: %r: %r", self.current_master, master_instance, master_node)
            self.current_master = master_instance
            if self.own_db and self.own_db != master_instance and self.config.get("autofollow"):
                self.start_following_new_master(master_instance)

        own_state = self.cluster_state.get(self.own_db)
        self.emit_stats(own_state)

        # If we're an observer ourselves, we'll grab the IP address from HTTP server address
        observer_info = ",".join(observer_state) or "no"
        if not self.own_db:
            observer_info = self.config.get("http_address", observer_info)
        standby_info = ",".join(standby_nodes) or "no"
        self.log.debug("Cluster has %s standbys, %s observers and %s as master, own_db: %r, own_state: %r",
                       standby_info, observer_info, self.current_master, self.own_db, own_state or "observer")

        if self.own_db:
            if self.own_db == self.current_master:
                # We are the master of this cluster, nothing to do
                self.log.debug("We %r: %r are still the master node: %r of this cluster, nothing to do.",
                               self.own_db, own_state, master_node)
                return
            if not standby_nodes:
                self.log.warning("No standby nodes set, master node: %r", master_node)
                return
            self.consider_failover(own_state, master_node, standby_nodes)

    def consider_failover(self, own_state, master_node, standby_nodes):
        if not master_node:
            # no master node at all in the cluster?
            self.log.warning("No master node in cluster, %r standby nodes exist, "
                             "%.2f seconds since last cluster config update, failover timeout set to %r seconds",
                             len(standby_nodes), time.time() - self.cluster_nodes_change_time,
                             self.replication_lag_failover_timeout)
            if self.current_master:
                self.trigger_check_queue.put("Master is missing, ask for immediate state check")
                if (time.time() - self.cluster_nodes_change_time) >= self.missing_master_from_config_timeout:
                    # we've seen a master at some point in time, but now it's
                    # missing, perform an immediate failover to promote one of
                    # the standbys
                    self.log.warning("Performing failover decision because existing master node "
                                     "disappeared from configuration")
                    self.do_failover_decision(own_state, standby_nodes)
                    return
            elif (time.time() - self.cluster_nodes_change_time) >= self.replication_lag_failover_timeout:
                # we've never seen a master and more than failover_timeout
                # seconds have passed since last config load (and start of
                # connection attempts to other nodes); perform failover
                self.log.warning("Performing failover decision because no master node was seen in cluster before timeout")
                self.do_failover_decision(own_state, standby_nodes)
                return
        self.check_replication_lag(own_state, standby_nodes)

    def check_replication_lag(self, own_state, standby_nodes):
        if self.is_restoring_or_catching_up_normally(own_state):
            # do not raise alerts during catchup at restore
            return

        replication_lag = own_state.get('replication_time_lag')
        if not replication_lag:
            self.log.warning("No replication lag set in own node state: %r", own_state)
            return
        if replication_lag >= self.replication_lag_warning_boundary:
            self.log.warning("Replication time lag has grown to: %r which is over WARNING boundary: %r, %r",
                             replication_lag, self.replication_lag_warning_boundary,
                             self.replication_lag_over_warning_limit)
            if not self.replication_lag_over_warning_limit:  # we just went over the boundary
                self.replication_lag_over_warning_limit = True
                self.create_alert_file("replication_delay_warning")
                if self.over_warning_limit_command:
                    self.log.warning("Executing over_warning_limit_command: %r", self.over_warning_limit_command)
                    return_code = self.execute_external_command(self.over_warning_limit_command)
                    self.log.warning("Executed over_warning_limit_command: %r, return_code: %r",
                                     self.over_warning_limit_command, return_code)
                else:
                    self.log.warning("No over_warning_limit_command set")
        elif self.replication_lag_over_warning_limit:
            self.replication_lag_over_warning_limit = False
            self.delete_alert_file("replication_delay_warning")

        if replication_lag >= self.replication_lag_failover_timeout:
            self.log.warning("Replication time lag has grown to: %r which is over CRITICAL boundary: %r"
                             ", checking if we need to failover",
                             replication_lag, self.replication_lag_failover_timeout)
            self.do_failover_decision(own_state, standby_nodes)
        else:
            self.log.debug("Replication lag was: %r, other nodes status was: %r", replication_lag, standby_nodes)

    def get_replication_positions(self, standby_nodes):
        self.log.debug("Getting replication positions from: %r", standby_nodes)
        known_replication_positions = {}
        for instance, node_state in standby_nodes.items():
            now = datetime.datetime.utcnow()
            if node_state['connection'] and \
                now - parse_iso_datetime(node_state['fetch_time']) < datetime.timedelta(seconds=20) and \
                instance not in self.never_promote_these_nodes:  # noqa # pylint: disable=line-too-long
                # use pg_last_xlog_receive_location if it's available,
                # otherwise fall back to pg_last_xlog_replay_location but
                # note that both of them can be None.  We prefer
                # receive_location over replay_location as some nodes may
                # not yet have replayed everything they've received, but
                # also consider the replay location in case receive_location
                # is empty as a node that has been brought up from backups
                # without ever connecting to a master will not have an empty
                # pg_last_xlog_receive_location
                lsn = node_state['pg_last_xlog_receive_location'] or node_state['pg_last_xlog_replay_location']
                xlog_pos = convert_xlog_location_to_offset(lsn) if lsn else 0
                known_replication_positions.setdefault(xlog_pos, set()).add(instance)
        return known_replication_positions

    def _been_in_contact_with_master_within_failover_timeout(self):
        # no need to do anything here if there are no disconnected masters
        if len(self.disconnected_master_nodes) > 0:
            disconnected_master_node = list(self.disconnected_master_nodes.values())[0]
            db_time = disconnected_master_node.get('db_time', get_iso_timestamp()) or get_iso_timestamp()
            time_since_last_contact = datetime.datetime.utcnow() - parse_iso_datetime(db_time)
            if time_since_last_contact < datetime.timedelta(seconds=self.replication_lag_failover_timeout):
                self.log.debug("We've had contact with master: %r at: %r within the last %.2fs, not failing over",
                               disconnected_master_node, db_time, time_since_last_contact.total_seconds())
                return True
        return False

    def do_failover_decision(self, own_state, standby_nodes):
        if len(self.connected_master_nodes) > 0 or self._been_in_contact_with_master_within_failover_timeout():
            self.log.warning("We still have some connected masters: %r, not failing over", self.connected_master_nodes)
            return

        known_replication_positions = self.get_replication_positions(standby_nodes)
        if not known_replication_positions:
            self.log.warning("No known replication positions, canceling failover consideration")
            return

        # If there are multiple nodes with the same replication positions pick the one with the "highest" name
        # to make sure pglookouts running on all standbys make the same decision.  The rationale for picking
        # the "highest" node is that there's no obvious way for pglookout to decide which of the nodes is
        # "best" beyond looking at replication positions, but picking the highest id supports environments
        # where nodes are assigned identifiers from an incrementing sequence identifiers and where we want to
        # promote the latest and greatest node.  In static environments node identifiers can be priority
        # numbers, with the highest number being the one that should be preferred.
        furthest_along_instance = max(known_replication_positions[max(known_replication_positions)])
        self.log.warning("Node that is furthest along is: %r, all replication positions were: %r",
                         furthest_along_instance, known_replication_positions)
        total_observers = len(self.connected_observer_nodes) + len(self.disconnected_observer_nodes)
        # +1 in the calculation comes from the master node
        total_amount_of_nodes = len(standby_nodes) + 1 - len(self.never_promote_these_nodes) + total_observers
        size_of_needed_majority = total_amount_of_nodes * 0.5
        amount_of_known_replication_positions = 0
        for known_replication_position in known_replication_positions.values():
            amount_of_known_replication_positions += len(known_replication_position)
        size_of_known_state = amount_of_known_replication_positions + len(self.connected_observer_nodes)
        self.log.debug("Size of known state: %.2f, needed majority: %r, %r/%r", size_of_known_state,
                       size_of_needed_majority, amount_of_known_replication_positions, int(total_amount_of_nodes))

        if standby_nodes[furthest_along_instance] == own_state:
            if self.check_for_maintenance_mode_file():
                self.log.warning("Canceling failover even though we were the node the furthest along, since "
                                 "this node has an existing maintenance_mode_file: %r",
                                 self.config.get("maintenance_mode_file", "/tmp/pglookout_maintenance_mode_file"))
                return
            elif self.own_db in self.never_promote_these_nodes:
                self.log.warning("Not doing a failover even though we were the node the furthest along, since this node: %r"
                                 " should never be promoted to master", self.own_db)
            elif size_of_known_state < size_of_needed_majority:
                self.log.warning("Not doing a failover even though we were the node the furthest along, since we aren't "
                                 "aware of the states of enough of the other nodes")
            else:
                start_time = time.time()
                self.log.warning("We will now do a failover to ourselves since we were the instance furthest along")
                return_code = self.execute_external_command(self.failover_command)
                self.log.warning("Executed failover command: %r, return_code: %r, took: %.2fs",
                                 self.failover_command, return_code, time.time() - start_time)
                self.create_alert_file("failover_has_happened")
                # Sleep for failover time to give the DB time to restart in promotion mode
                # You want to use this if the failover command is not one that blocks until
                # the db has restarted
                time.sleep(self.config.get("failover_sleep_time", 0.0))
                if return_code == 0:
                    self.replication_lag_over_warning_limit = False
                    self.delete_alert_file("replication_delay_warning")
        else:
            self.log.warning("Nothing to do since node: %r is the furthest along", furthest_along_instance)

    def modify_recovery_conf_to_point_at_new_master(self, new_master_instance):
        path_to_recovery_conf = os.path.join(self.config.get("pg_data_directory"), "recovery.conf")
        with open(path_to_recovery_conf, "r") as fp:
            old_conf = fp.read().splitlines()
        has_recovery_target_timeline = False
        new_conf = []
        old_conn_info = None
        for line in old_conf:
            if line.startswith("recovery_target_timeline"):
                has_recovery_target_timeline = True
            if line.startswith("primary_conninfo"):
                # grab previous entry: strip surrounding quotes and replace two quotes with one
                try:
                    old_conn_info = get_connection_info_from_config_line(line)
                except ValueError:
                    self.log.exception("failed to parse previous %r, ignoring", line)
                continue  # skip this line
            new_conf.append(line)

        # If has_recovery_target_timeline is set and old_conn_info matches
        # new info we don't have to do anything
        new_conn_info = get_connection_info(self.primary_conninfo_template)
        master_instance_conn_info = get_connection_info(self.config["remote_conns"][new_master_instance])
        assert "host" in master_instance_conn_info
        new_conn_info["host"] = master_instance_conn_info["host"]
        if "port" in master_instance_conn_info:
            new_conn_info["port"] = master_instance_conn_info["port"]
        if new_conn_info == old_conn_info and has_recovery_target_timeline:
            self.log.debug("recovery.conf already contains conninfo matching %r, not updating", new_master_instance)
            return False
        # Otherwise we append the new primary_conninfo
        new_conf.append("primary_conninfo = {0}".format(adapt(create_connection_string(new_conn_info))))
        # The timeline of the recovery.conf will require a higher timeline target
        if not has_recovery_target_timeline:
            new_conf.append("recovery_target_timeline = 'latest'")
        # prepend our tag
        new_conf.insert(0,
                        "# pglookout updated primary_conninfo for instance {0} at {1}"
                        .format(new_master_instance, get_iso_timestamp()))
        # Replace old recovery.conf with a fresh copy
        with open(path_to_recovery_conf + "_temp", "w") as fp:
            fp.write("\n".join(new_conf) + "\n")
        self.log.debug("Previous recovery.conf: %s", old_conf)
        self.log.debug("Newly written recovery.conf: %s", new_conf)
        os.rename(path_to_recovery_conf + "_temp", path_to_recovery_conf)
        return True

    def start_following_new_master(self, new_master_instance):
        start_time = time.time()
        updated_config = self.modify_recovery_conf_to_point_at_new_master(new_master_instance)
        if not updated_config:
            self.log.info("Already following master %r, no need to start following it again", new_master_instance)
            return
        start_command = self.config.get("pg_start_command", "").split()
        stop_command = self.config.get("pg_stop_command", "").split()
        self.log.info("Starting to follow new master %r, modified recovery.conf and restarting PostgreSQL"
                      "; pg_stop_command %r; pg_start_command %r",
                      new_master_instance, start_command, stop_command)
        self.execute_external_command(stop_command)
        self.execute_external_command(start_command)
        self.log.info("Started following new master %r, took: %.2fs", new_master_instance, time.time() - start_time)

    def execute_external_command(self, command):
        self.log.warning("Executing external command: %r", command)
        return_code, output = 0, ""
        try:
            output = subprocess.check_call(command)
        except subprocess.CalledProcessError as err:
            self.log.exception("Problem with executing: %r, return_code: %r, output: %r",
                               command, err.returncode, err.output)
            self.stats.unexpected_exception(err, where="execute_external_command")
            return_code = err.returncode  # pylint: disable=no-member
        self.log.warning("Executed external command: %r, output: %r", return_code, output)
        return return_code

    def check_for_maintenance_mode_file(self):
        return os.path.exists(self.config.get("maintenance_mode_file", "/tmp/pglookout_maintenance_mode_file"))

    def create_alert_file(self, filename):
        try:
            filepath = os.path.join(self.config.get("alert_file_dir", os.getcwd()), filename)
            self.log.debug("Creating alert file: %r", filepath)
            with open(filepath, "w") as fp:
                fp.write("alert")
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Problem writing alert file: %r", filepath)
            self.stats.unexpected_exception(ex, where="create_alert_file")

    def delete_alert_file(self, filename):
        try:
            filepath = os.path.join(self.config.get("alert_file_dir", os.getcwd()), filename)
            if os.path.exists(filepath):
                self.log.debug("Deleting alert file: %r", filepath)
                os.unlink(filepath)
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Problem unlinking: %r", filepath)
            self.stats.unexpected_exception(ex, where="delete_alert_file")

    def main_loop(self):
        while self.running:
            # Separate try/except so we still write the state file
            sleep_time = 5.0
            try:
                sleep_time = float(self.config.get("replication_state_check_interval", 5.0))
                self.check_cluster_state()
            except Exception as ex:  # pylint: disable=broad-except
                self.log.exception("Failed to check cluster state")
                self.stats.unexpected_exception(ex, where="main_loop1")
            try:
                self.write_cluster_state_to_json_file()
            except Exception as ex:  # pylint: disable=broad-except
                self.log.exception("Failed to write cluster state")
                self.stats.unexpected_exception(ex, where="main_loop2")
            time.sleep(sleep_time)

    def run(self):
        self.cluster_monitor.start()
        self.webserver.start()
        self.main_loop()


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog="pglookout",
        description="postgresql replication monitoring and failover daemon")
    parser.add_argument("--version", action="version", help="show program version",
                        version=version.__version__)
    parser.add_argument("config", help="configuration file")
    arg = parser.parse_args(args)

    if not os.path.exists(arg.config):
        print("pglookout: {!r} doesn't exist".format(arg.config))
        return 1

    logutil.configure_logging()

    pglookout = PgLookout(arg.config)
    pglookout.run()


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""
pglookout - cluster monitoring component

Copyright (c) 2015 Ohmu Ltd
Copyright (c) 2014 F-Secure

This file is under the Apache License, Version 2.0.
See the file `LICENSE` for details.
"""

from . import logutil
from .common import get_iso_timestamp, parse_iso_datetime
from .pgutil import mask_connection_info
from concurrent.futures import as_completed, ThreadPoolExecutor
from email.utils import parsedate
from psycopg2.extras import RealDictCursor
from threading import Thread
import datetime
import errno
import logging
import psycopg2
import requests
import select
import time

try:
    from queue import Empty  # pylint: disable=import-error
except ImportError:
    from Queue import Empty  # pylint: disable=import-error


class PglookoutTimeout(Exception):
    pass


def wait_select(conn, timeout=5.0):
    end_time = time.time() + timeout
    while time.time() < end_time:
        time_left = end_time - time.time()
        state = conn.poll()
        try:
            if state == psycopg2.extensions.POLL_OK:
                return
            elif state == psycopg2.extensions.POLL_READ:
                select.select([conn.fileno()], [], [], min(timeout, time_left))
            elif state == psycopg2.extensions.POLL_WRITE:
                select.select([], [conn.fileno()], [], min(timeout, time_left))
            else:
                raise psycopg2.OperationalError("bad state from poll: %s" % state)
        except select.error as error:
            if error.args[0] != errno.EINTR:
                raise
    raise PglookoutTimeout("timed out in wait_select")


class ClusterMonitor(Thread):
    def __init__(self, config, cluster_state, observer_state, create_alert_file, trigger_check_queue,
                 stats):
        Thread.__init__(self)
        self.log = logging.getLogger("ClusterMonitor")
        self.stats = stats
        self.running = True
        self.cluster_state = cluster_state
        self.observer_state = observer_state
        self.config = config
        self.create_alert_file = create_alert_file
        self.db_conns = {}
        self.trigger_check_queue = trigger_check_queue
        self.session = requests.Session()
        if self.config.get("syslog"):
            self.syslog_handler = logutil.set_syslog_handler(
                address=self.config.get("syslog_address", "/dev/log"),
                facility=self.config.get("syslog_facility", "local2"),
                logger=self.log,
            )
        self.log.debug("Initialized ClusterMonitor with: %r", cluster_state)

    def _connect_to_db(self, instance, dsn):
        conn = self.db_conns.get(instance)
        if conn:
            return conn
        if not dsn:
            self.log.warning("Can't connect to %s, dsn is %r", instance, dsn)
            return None
        inst_info_str = "{0!r} ({1})".format(instance, mask_connection_info(dsn))
        try:
            self.log.info("Connecting to %s", inst_info_str)
            conn = psycopg2.connect(dsn=dsn, async=True)
            wait_select(conn)
            self.log.debug("Connected to %s", inst_info_str)
        except (PglookoutTimeout, psycopg2.OperationalError) as ex:
            self.log.warning("%s (%s) connecting to %s (%s)",
                             ex.__class__.__name__, str(ex).strip(),
                             instance, inst_info_str)
            if "password authentication" in getattr(ex, "message", ""):
                self.create_alert_file("authentication_error")
            conn = None  # make sure we don't try to use the connection if we timed out
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Failed to connect to %s (%s)",
                               instance, inst_info_str)
            self.stats.unexpected_exception(ex, where="_connect_to_db")
            conn = None
        self.db_conns[instance] = conn
        return conn

    def _fetch_observer_state(self, instance, uri):
        result = {"fetch_time": get_iso_timestamp(), "connection": True}
        try:
            fetch_uri = uri + "/state.json"
            response = self.session.get(fetch_uri, timeout=5.0)

            # check time difference for large skews
            remote_server_time = parsedate(response.headers['date'])
            remote_server_time = datetime.datetime.fromtimestamp(time.mktime(remote_server_time))
            time_diff = parse_iso_datetime(result['fetch_time']) - remote_server_time
            if time_diff > datetime.timedelta(seconds=5):
                self.log.error("Time difference between us and observer node %r is %r, response: %r, ignoring response",
                               instance, time_diff, response.json())  # pylint: disable=no-member
                return
            result.update(response.json())  # pylint: disable=no-member
        except requests.ConnectionError as ex:
            self.log.warning("%s (%s) fetching state from observer: %r, %r",
                             ex.__class__.__name__, ex, instance, fetch_uri)
            result['connection'] = False
        except Exception as ex:  # pylint: disable=broad-except
            self.log.exception("Problem in fetching state from observer: %r, %r", instance, fetch_uri)
            self.stats.unexpected_exception(ex, where="_fetch_observer_state")
            result['connection'] = False
        return result

    def fetch_observer_state(self, instance, uri):
        start_time = time.time()
        result = self._fetch_observer_state(instance, uri)
        if result:
            if instance in self.observer_state:
                self.observer_state[instance].update(result)
            else:
                self.observer_state[instance] = result
        self.log.debug("Observer: %r state was: %r, took: %.4fs to fetch",
                       instance, result, time.time() - start_time)

    def connect_to_cluster_nodes_and_cleanup_old_nodes(self):
        leftover_conns = set(self.db_conns) - set(self.config.get("remote_conns", {}))
        for leftover_instance in leftover_conns:
            self.log.debug("Removing leftover state for: %r", leftover_instance)
            self.db_conns.pop(leftover_instance)
            self.cluster_state.pop(leftover_instance, "")
            self.observer_state.pop(leftover_instance, "")
        #  Making sure we have a connection to all currently configured db hosts
        for instance, connect_string in self.config.get("remote_conns", {}).items():
            self._connect_to_db(instance, dsn=connect_string)

    def _standby_status_query(self, instance, db_conn):
        """Status query that is executed on the standby node"""
        f_result = None
        result = {"fetch_time": get_iso_timestamp(), "connection": False}
        if not db_conn:
            db_conn = self._connect_to_db(instance, self.config["remote_conns"].get(instance))
            if not db_conn:
                return result
        try:
            phase = "querying status from"
            self.log.debug("%s %r", phase, instance)
            c = db_conn.cursor(cursor_factory=RealDictCursor)
            fields = [
                "now() AS db_time",
                "pg_is_in_recovery()",
                "pg_last_xact_replay_timestamp()",
                "pg_last_xlog_receive_location()",
                "pg_last_xlog_replay_location()",
            ]
            query = "SELECT {}".format(", ".join(fields))
            c.execute(query)
            wait_select(c.connection)
            f_result = c.fetchone()
            if not f_result['pg_is_in_recovery']:
                # This is only run on masters to create txid traffic every db_poll_interval
                phase = "updating transaction on"
                self.log.debug("%s %r", phase, instance)
                c.execute("SELECT txid_current()")
                wait_select(c.connection)
        except (PglookoutTimeout, psycopg2.DatabaseError, psycopg2.InterfaceError, psycopg2.OperationalError) as ex:
            self.log.warning("%s (%s) %s %s", ex.__class__.__name__, str(ex).strip(), phase, instance)
            db_conn.close()
            self.db_conns[instance] = None

        result.update(self._parse_status_query_result(f_result))
        return result

    @staticmethod
    def _parse_status_query_result(result):
        if not result:
            return {}
        # abs is for catching time travel (as in going from the future to the past
        if result["pg_last_xact_replay_timestamp"]:
            replication_time_lag = abs(result["db_time"] - result["pg_last_xact_replay_timestamp"])
            result["replication_time_lag"] = replication_time_lag.total_seconds()
            result["pg_last_xact_replay_timestamp"] = get_iso_timestamp(result["pg_last_xact_replay_timestamp"])

        if not result["pg_is_in_recovery"]:
            # These are set to None so when we query a standby promoted to master
            # it looks identical to the results from a master node that's never been a standby
            result.update({
                "pg_last_xlog_receive_location": None,
                "pg_last_xact_replay_timestamp": None,
                "pg_last_xlog_replay_location": None,
                "replication_time_lag": None,  # differentiate from actual lag=0.0
            })
        result.update({"db_time": get_iso_timestamp(result["db_time"]), "connection": True})
        return result

    def standby_status_query(self, instance, db_conn):
        start_time = time.time()
        result = self._standby_status_query(instance, db_conn)
        self.log.debug("DB state gotten from: %r was: %r, took: %.4fs to fetch",
                       instance, result, time.time() - start_time)
        if instance in self.cluster_state:
            self.cluster_state[instance].update(result)
        else:
            self.cluster_state[instance] = result

        # record the first time we saw replication happen from the master
        if result.get("pg_last_xlog_receive_location"):
            result.setdefault("replication_start_time", time.time())

        # maintain lowest seen lag in seconds in the state
        min_lag = self.cluster_state[instance].get("min_replication_time_lag")
        now_lag = result.get("replication_time_lag")
        if now_lag is not None:
            if min_lag is None:
                self.cluster_state[instance]["min_replication_time_lag"] = now_lag
            else:
                self.cluster_state[instance]["min_replication_time_lag"] = min(min_lag, now_lag)

    def main_monitoring_loop(self):
        self.connect_to_cluster_nodes_and_cleanup_old_nodes()
        thread_count = len(self.db_conns) + len(self.config.get("observers", {}))
        futures = []
        with ThreadPoolExecutor(max_workers=thread_count) as tex:
            for instance, db_conn in self.db_conns.items():
                futures.append(tex.submit(self.standby_status_query, instance, db_conn))
            for instance, uri in self.config.get("observers", {}).items():
                futures.append(tex.submit(self.fetch_observer_state, instance, uri))
            for future in as_completed(futures):
                if future.exception():
                    self.log.error("Got error: %r when checking cluster state", future.exception())

    def run(self):
        while self.running:
            self.main_monitoring_loop()
            try:
                self.trigger_check_queue.get(timeout=self.config.get("db_poll_interval", 5.0))
            except Empty:
                pass



# Copied from https://github.com/ohmu/ohmu_common_py version.py version 0.0.1-0-unknown-fa54b44
"""
pglookout - version detection and version.py __version__ generation

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

import imp
import os
import subprocess


def save_version(new_ver, old_ver, version_file):
    if not new_ver:
        return False
    version_file = os.path.join(os.path.dirname(__file__), version_file)
    if not old_ver or new_ver != old_ver:
        with open(version_file, "w") as fp:
            fp.write("__version__ = '{}'\n".format(new_ver))
    return True


def get_project_version(version_file):
    version_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), version_file)
    try:
        module = imp.load_source("verfile", version_file)
        file_ver = module.__version__
    except IOError:
        file_ver = None

    os.chdir(os.path.dirname(__file__) or ".")
    try:
        git_out = subprocess.check_output(["git", "describe", "--always"],
                                          stderr=getattr(subprocess, "DEVNULL", None))
    except (OSError, subprocess.CalledProcessError):
        pass
    else:
        git_ver = git_out.splitlines()[0].strip().decode("utf-8")
        if "." not in git_ver:
            git_ver = "0.0.1-0-unknown-{}".format(git_ver)
        if save_version(git_ver, file_ver, version_file):
            return git_ver

    makefile = os.path.join(os.path.dirname(__file__), "Makefile")
    if os.path.exists(makefile):
        with open(makefile, "r") as fp:
            lines = fp.readlines()
        short_ver = [line.split("=", 1)[1].strip() for line in lines if line.startswith("short_ver")][0]
        if save_version(short_ver, file_ver, version_file):
            return short_ver

    if not file_ver:
        raise Exception("version not available from git or from file {!r}".format(version_file))

    return file_ver

if __name__ == "__main__":
    import sys
    get_project_version(sys.argv[1])



"""
pglookout - common utility functions

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""
import datetime
import re


def convert_xlog_location_to_offset(xlog_location):
    log_id, offset = xlog_location.split("/")
    return int(log_id, 16) << 32 | int(offset, 16)


ISO_EXT_RE = re.compile(r'(?P<year>\d{4})-(?P<month>\d\d)-(?P<day>\d\d)(T(?P<hour>\d\d):(?P<minute>\d\d)'
                        r'(:(?P<second>\d\d)(.(?P<microsecond>\d{6}))?)?Z)?$')
ISO_BASIC_RE = re.compile(r'(?P<year>\d{4})(?P<month>\d\d)(?P<day>\d\d)(T(?P<hour>\d\d)(?P<minute>\d\d)'
                          r'((?P<second>\d\d)((?P<microsecond>\d{6}))?)?Z)?$')


def parse_iso_datetime(value):
    match = ISO_EXT_RE.match(value)
    if not match:
        match = ISO_BASIC_RE.match(value)
    if not match:
        raise ValueError("Invalid ISO timestamp {0!r}".format(value))
    parts = dict((key, int(match.group(key) or '0'))
                 for key in ('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond'))
    return datetime.datetime(tzinfo=None, **parts)


def get_iso_timestamp(fetch_time=None):
    if not fetch_time:
        fetch_time = datetime.datetime.utcnow()
    elif fetch_time.tzinfo:
        fetch_time = fetch_time.replace(tzinfo=None) - datetime.timedelta(seconds=fetch_time.utcoffset().seconds)
    return fetch_time.isoformat() + "Z"



# Copied from https://github.com/ohmu/ohmu_common_py ohmu_common_py/pgutil.py version 0.0.1-0-unknown-fa54b44
"""
pglookout - postgresql utility functions

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

try:
    from urllib.parse import urlparse, parse_qs  # pylint: disable=no-name-in-module, import-error
except ImportError:
    from urlparse import urlparse, parse_qs  # pylint: disable=no-name-in-module, import-error


def create_connection_string(connection_info):
    return " ".join("{}='{}'".format(k, str(v).replace("'", "\\'"))
                    for k, v in sorted(connection_info.items()))


def mask_connection_info(info):
    masked_info = get_connection_info(info)
    password = masked_info.pop("password", None)
    return "{0}; {1} password".format(
        create_connection_string(masked_info),
        "no" if password is None else "hidden")


def get_connection_info_from_config_line(line):
    _, value = line.split("=", 1)
    value = value.strip()[1:-1].replace("''", "'")
    return get_connection_info(value)


def get_connection_info(info):
    """turn a connection info object into a dict or return it if it was a
    dict already.  supports both the traditional libpq format and the new
    url format"""
    if isinstance(info, dict):
        return info.copy()
    elif info.startswith("postgres://") or info.startswith("postgresql://"):
        return parse_connection_string_url(info)
    else:
        return parse_connection_string_libpq(info)


def parse_connection_string_url(url):
    # drop scheme from the url as some versions of urlparse don't handle
    # query and path properly for urls with a non-http scheme
    schemeless_url = url.split(":", 1)[1]
    p = urlparse(schemeless_url)
    fields = {}
    if p.hostname:
        fields["host"] = p.hostname
    if p.port:
        fields["port"] = str(p.port)
    if p.username:
        fields["user"] = p.username
    if p.password is not None:
        fields["password"] = p.password
    if p.path and p.path != "/":
        fields["dbname"] = p.path[1:]
    for k, v in parse_qs(p.query).items():
        fields[k] = v[-1]
    return fields


def parse_connection_string_libpq(connection_string):
    """parse a postgresql connection string as defined in
    http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-CONNSTRING"""
    fields = {}
    while True:
        connection_string = connection_string.strip()
        if not connection_string:
            break
        if "=" not in connection_string:
            raise ValueError("expecting key=value format in connection_string fragment {!r}".format(connection_string))
        key, rem = connection_string.split("=", 1)
        if rem.startswith("'"):
            asis, value = False, ""
            for i in range(1, len(rem)):
                if asis:
                    value += rem[i]
                    asis = False
                elif rem[i] == "'":
                    break  # end of entry
                elif rem[i] == "\\":
                    asis = True
                else:
                    value += rem[i]
            else:
                raise ValueError("invalid connection_string fragment {!r}".format(rem))
            connection_string = rem[i + 1:]  # pylint: disable=undefined-loop-variable
        else:
            res = rem.split(None, 1)
            if len(res) > 1:
                value, connection_string = res
            else:
                value, connection_string = rem, ""
        fields[key] = value
    return fields



"""
pglookout

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

from pglookout.common import (
    convert_xlog_location_to_offset,
    parse_iso_datetime, get_iso_timestamp, ISO_EXT_RE,
)
from pytest import raises
import datetime


def test_convert_xlog_location_to_offset():
    assert convert_xlog_location_to_offset("1/00000000") == 1 << 32
    assert convert_xlog_location_to_offset("F/AAAAAAAA") == (0xF << 32) | 0xAAAAAAAA
    with raises(ValueError):
        convert_xlog_location_to_offset("x")
    with raises(ValueError):
        convert_xlog_location_to_offset("x/y")


def test_parse_iso_datetime():
    date = datetime.datetime.utcnow()
    date.replace(microsecond=0)
    assert date == parse_iso_datetime(date.isoformat() + "Z")
    with raises(ValueError):
        parse_iso_datetime("foobar")


def test_get_iso_timestamp():
    v = get_iso_timestamp()
    assert ISO_EXT_RE.match(v)
    ts = datetime.datetime.now()
    v = get_iso_timestamp(ts)
    assert parse_iso_datetime(v) == ts



"""
pglookout

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

from pglookout import statsd
from pglookout.cluster_monitor import ClusterMonitor
from psycopg2.extensions import POLL_OK
from datetime import datetime, timedelta

try:
    from queue import Queue  # pylint: disable=import-error
    from unittest.mock import MagicMock, Mock, patch  # pylint: disable=no-name-in-module
except ImportError:
    from Queue import Queue  # pylint: disable=import-error
    from mock import MagicMock, Mock, patch  # pylint: disable=import-error


def test_replication_lag():
    # pylint: disable=protected-access
    now = datetime.now()
    status = {
        "db_time": now,
        "pg_is_in_recovery": True,
        "pg_last_xact_replay_timestamp": now,
        "pg_last_xlog_receive_location": "0/0000001",
        "pg_last_xlog_replay_location": "0/0000002",
    }
    result = ClusterMonitor._parse_status_query_result(status.copy())
    assert result["replication_time_lag"] == 0.0
    status["db_time"] += timedelta(seconds=50, microseconds=42)
    result = ClusterMonitor._parse_status_query_result(status.copy())
    assert result["replication_time_lag"] == 50.000042
    status["db_time"] = now + timedelta(hours=42)
    result = ClusterMonitor._parse_status_query_result(status.copy())
    assert result["replication_time_lag"] == 151200.0


@patch("psycopg2.connect")
def test_main_loop(psycopg2_connect):
    config = {"remote_conns":
              {"foo": "host=1.2.3.4 dbname=postgres user=pglookout password=fake_pass",
               "bar": "host=2.3.4.5 dbname=postgres user=pglookout password=fake_pass"}}
    cluster_state = {}
    observer_state = {}

    def alert_file_func(arg):  # pylint: disable=unused-argument
        pass

    create_alert_file = alert_file_func
    trigger_check_queue = Queue()
    trigger_check_queue.put("test entry so we don't wait five seconds to get one")

    class FakeCursor(MagicMock):  # pylint: disable=too-many-ancestors
        def execute(self, query):  # pylint: disable=no-self-use,unused-argument
            return

        def fetchone(self):  # pylint: disable=no-self-use
            return {"pg_is_in_recovery": False, "pg_last_xact_replay_timestamp": datetime.utcnow(),
                    "db_time": datetime.utcnow()}

    class FakeConn(Mock):  # pylint: disable=too-many-ancestors
        def cursor(self, cursor_factory):  # pylint: disable=unused-argument
            f = FakeCursor()
            f.connection = self  # pylint: disable=attribute-defined-outside-init
            return f

        def poll(self):  # pylint: disable=no-self-use
            return POLL_OK
    psycopg2_connect.return_value = FakeConn()

    cm = ClusterMonitor(
        config=config,
        cluster_state=cluster_state,
        observer_state=observer_state,
        create_alert_file=create_alert_file,
        trigger_check_queue=trigger_check_queue,
        stats=statsd.StatsClient(host=None))
    cm.main_monitoring_loop()

    assert len(cm.cluster_state) == 2
    assert "foo" in cm.cluster_state
    assert "bar" in cm.cluster_state



# Copied from https://github.com/ohmu/ohmu_common_py test/test_pgutil.py version 0.0.1-0-unknown-fa54b44
"""
pglookout - postgresql utility function tests

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

from pglookout.pgutil import (
    create_connection_string, get_connection_info, mask_connection_info,
)
from pytest import raises


def test_connection_info():
    url = "postgres://hannu:secret@dbhost.local:5555/abc?replication=true&sslmode=foobar&sslmode=require"
    cs = "host=dbhost.local user='hannu'   dbname='abc'\n" \
         "replication=true   password=secret sslmode=require port=5555"
    ci = {
        "host": "dbhost.local",
        "port": "5555",
        "user": "hannu",
        "password": "secret",
        "dbname": "abc",
        "replication": "true",
        "sslmode": "require",
    }
    assert get_connection_info(ci) == get_connection_info(cs)
    assert get_connection_info(ci) == get_connection_info(url)

    basic_cstr = "host=localhost user=os"
    assert create_connection_string(get_connection_info(basic_cstr)) == "host='localhost' user='os'"

    assert get_connection_info("foo=bar bar='\\'x'") == {"foo": "bar", "bar": "'x"}

    with raises(ValueError):
        get_connection_info("foo=bar x")
    with raises(ValueError):
        get_connection_info("foo=bar bar='x")


def test_mask_connection_info():
    url = "postgres://michael:secret@dbhost.local:5555/abc?replication=true&sslmode=foobar&sslmode=require"
    cs = "host=dbhost.local user='michael'   dbname='abc'\n" \
         "replication=true   password=secret sslmode=require port=5555"
    ci = get_connection_info(cs)
    masked_url = mask_connection_info(url)
    masked_cs = mask_connection_info(url)
    masked_ci = mask_connection_info(url)
    assert masked_url == masked_cs
    assert masked_url == masked_ci
    assert "password" in ci  # make sure we didn't modify the original dict

    # the return format is a connection string without password, followed by
    # a semicolon and comment about password presence
    masked_str, password_info = masked_url.split("; ", 1)
    assert "password" not in masked_str
    assert password_info == "hidden password"

    # remasking the masked string should yield a no password comment
    masked_masked = mask_connection_info(masked_str)
    _, masked_password_info = masked_masked.split("; ", 1)
    assert masked_password_info == "no password"



"""
pglookout_current_master - display the current cluster master

Copyright (c) 2015 Ohmu Ltd
Copyright (c) 2014 F-Secure

This file is under the Apache License, Version 2.0.
See the file `LICENSE` for details.
"""

from __future__ import print_function
from . import version
import argparse
import json
import os
import sys
import time


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog="pglookout_current_master",
        description="postgresql replication monitoring and failover daemon")
    parser.add_argument("--version", action="version", help="show program version",
                        version=version.__version__)
    parser.add_argument("state", help="pglookout state file")
    arg = parser.parse_args(args)

    if not os.path.exists(arg.state):
        print("pglookout_current_master: {!r} doesn't exist".format(arg.state))
        return 1

    try:
        with open(arg.state, "r") as fp:
            config = json.load(fp)
        state_file_path = config.get("json_state_file_path", "/tmp/pglookout_state.json")  # pylint: disable=no-member
        if time.time() - os.stat(state_file_path).st_mtime > 60.0:
            # file older than one minute, pglookout probably dead, exit with minus one
            return -1
        with open(state_file_path, "r") as fp:
            state_dict = json.load(fp)
        current_master = state_dict['current_master']
        print(current_master)
    except:  # pylint: disable=bare-except
        return -1


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copied from https://github.com/ohmu/ohmu_common_py ohmu_common_py/logutil.py version 0.0.1-0-unknown-fa54b44
"""
pglookout - logging formats and utility functions

Copyright (c) 2015 Ohmu Ltd
See LICENSE for details
"""

import logging
import logging.handlers
import os

try:
    from systemd import daemon  # pylint: disable=no-name-in-module
except ImportError:
    daemon = None


LOG_FORMAT = "%(asctime)s\t%(name)s\t%(threadName)s\t%(levelname)s\t%(message)s"
LOG_FORMAT_SHORT = "%(levelname)s\t%(message)s"
LOG_FORMAT_SYSLOG = "%(name)s %(threadName)s %(levelname)s: %(message)s"


def set_syslog_handler(address, facility, logger):
    syslog_handler = logging.handlers.SysLogHandler(address=address, facility=facility)
    logger.addHandler(syslog_handler)
    formatter = logging.Formatter(LOG_FORMAT_SYSLOG)
    syslog_handler.setFormatter(formatter)
    return syslog_handler


def configure_logging(level=logging.DEBUG, short_log=False):
    # Are we running under systemd?
    if os.getenv("NOTIFY_SOCKET"):
        logging.basicConfig(level=level, format=LOG_FORMAT_SYSLOG)
        if not daemon:
            print(
                "WARNING: Running under systemd but python-systemd not available, "
                "systemd won't see our notifications"
            )
    else:
        logging.basicConfig(level=level, format=LOG_FORMAT_SHORT if short_log else LOG_FORMAT)


def notify_systemd(status):
    if daemon:
        daemon.notify(status)



from setuptools import setup, find_packages
import os
import sys
import version


readme_path = os.path.join(os.path.dirname(__file__), "README.rst")
with open(readme_path, "r") as fp:
    readme_text = fp.read()


version_for_setup_py = version.get_project_version("pglookout/version.py")
version_for_setup_py = ".dev".join(version_for_setup_py.split("-", 2)[:2])


requires = [
    "psycopg2 >= 2.0.0",
    "requests >= 1.2.0",
]

if sys.version_info[0] == 2:
    requires.append("futures")


setup(
    name="pglookout",
    version=version_for_setup_py,
    zip_safe=False,
    packages=find_packages(exclude=["test"]),
    install_requires=requires,
    extras_require={},
    dependency_links=[],
    package_data={},
    data_files=[],
    entry_points={
        "console_scripts": [
            "pglookout = pglookout.pglookout:main",
            "pglookout_current_master = pglookout.current_master:main",
        ],
    },
    author="Hannu Valtonen",
    author_email="hannu.valtonen@ohmu.fi",
    license="Apache 2.0",
    platforms=["POSIX", "MacOS"],
    description="PostgreSQL replication monitoring and failover daemon",
    long_description=readme_text,
    url="https://github.com/ohmu/pglookout/",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Software Development :: Libraries",
    ],
)
